#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include "hash_db.h"

#ifndef HASH_ALLOC
#define ALLOC_ONE_HASH_E() (calloc(1, sizeof(HASH_ENTRY)))
#define FREE_HASH_E(X)      free(X)
#else
#include "hash_alloc.h"
#define ALLOC_ONE_HASH_E() (hash_alloc())
#define FREE_HASH_E(X)      hash_free(X)
#endif

static unsigned long primes[] = { 7, 11, 19, 31, 67, 131, 257, 509, 1021, 
				  2039, 4099, 8191, 16381, 32771, 65539,
				  131071, 262147, 524287, 1048573, 2097143,
				  4194301, 8388593, 16777213, 33554393,
				  67108859, 0 };

/*
 * The following hash algorithm is due to Ian Dall, with slight modifications
 * by Geoff Kuenning to reflect the results of testing with the English
 * dictionaries actually distributed with ispell.
 */
#define HASHSHIFT   5

int hash (const char *s, int keysize, int hashtblsize)
{
  register long       h = 0;
  register int        i;

  for (i = 0;  i < keysize  && i < 4; i++ )
        h = (h << 8) | s[i];

  for (; i < keysize; i++) {
    /*
     * We have to do circular shifts the hard way, since C doesn't
     * have them even though the hardware probably does.  Oh, well.
     */
    h = (h << HASHSHIFT)
      | ((h >> (32 - HASHSHIFT)) & ((1 << HASHSHIFT) - 1));
    h ^= s[i];
  }
  return (unsigned long) h % hashtblsize;
}

HASH_DB *hash_create(char *name, int size, int thres100) {
  HASH_DB *ret;
  if ((ret = calloc(1, sizeof(HASH_DB))) == NULL ||
      (ret->hdb_array = calloc(size, sizeof(HASH_ENTRY))) == NULL ||
      (ret->hdb_name = strdup(name)) == NULL) {
    perror("malloc (hash_create)");
    return NULL;
  }
  ret->hdb_size = size;
  ret->hdb_index = 1;
  ret->hdb_thres100 = thres100;
  MD5Init(&(ret->hdb_md5ctx));
  return ret;
}

static void rescale (HASH_DB *hash_db) {
  unsigned long i;
  unsigned long hval;
  unsigned long oldsize = hash_db->hdb_size;
  unsigned long newsize = 0;
  HASH_ENTRY *new_array;
  HASH_ENTRY *old_array = hash_db->hdb_array;

  for (i = 0; primes[i] != 0; i++)
    if ((hash_db->hdb_index / hash_db->hdb_thres100) * HASH_TRESH_SCALE < 
	primes[i]) {
      newsize = primes[i];
      break;
    }

  if (newsize == 0) {
      fprintf (stderr, "Unable to rescale hash %s (%p) from %lu to larger\n", 
	       hash_db->hdb_name, hash_db, oldsize);
      return;
  }

  fprintf (stderr, "Rescaling hash %s (%p) from %lu to ", 
	   hash_db->hdb_name, hash_db, oldsize);
  if ((new_array = calloc(newsize, sizeof(HASH_ENTRY))) == NULL) {
    perror("malloc (rescale)"); /* ok, nothing is lost.. */
    hash_db->hdb_thres100 += hash_db->hdb_thres100 / 10;
				/* increment threshold to avoid too
				 * frequent rescale overhead... */
    return;			
  }
  hash_db->hdb_array = new_array;
  hash_db->hdb_size = newsize;
  for (i = 0; i < oldsize; i++) { /* go through all */
    HASH_ENTRY *he, *oe;
    for (oe = &(old_array[i]); oe != NULL && oe->id != 0;) {
      hval = hash (oe->key, oe->keylen, hash_db->hdb_size);
      he = &(hash_db->hdb_array[hval]);
      if (he->id != 0) {		/* ok, there was something */
	HASH_ENTRY *hen = he;
	do {
	  he = hen;			/* store old */
	} while ((hen = hen->next));
	/* not found, but in linked list */
	if ((he->next = ALLOC_ONE_HASH_E()) == NULL) {
	  perror("calloc (fatal!)");
	  exit(10);
	};
	he = he->next;
      }
      he->id = oe->id;
      he->keylen = oe->keylen;
      memcpy (he->key, oe->key, oe->keylen);
      if (oe != &(old_array[i])) {
	HASH_ENTRY *tmp = oe;
	oe=oe->next;
	FREE_HASH_E (tmp);		/* free it */
      } else {
	oe=oe->next;
      }
    }
  }
  free(old_array);		/* free old array */
  fprintf (stderr, "%u slots (%d stored)\n", 
	   hash_db->hdb_size, hash_db->hdb_index);
}
const char *md5hash (HASH_DB *hash_db, const char *key, int *keylen) {
  static char digest[MD5LEN];

  if (*keylen > MD5LEN) {
    MD5_CTX md5 = hash_db->hdb_md5ctx;
    MD5Update (&md5, key, *keylen);
    MD5Final (digest, &md5);
    *keylen = MD5LEN;
    return digest;
  } else {
    return key;
  }
}

#define STRBUF 1024
int hash_put_print (HASH_DB *hash_db, const char *key,  int keylen) {
  int ind = hash_db->hdb_index;
  int i;
  if ((i = hash_put (hash_db, key,  keylen)) >= ind) {
    char s[STRBUF];
    size_t max = (keylen > STRBUF - 2)? STRBUF - 2 : keylen;
    strncpy (s, key, max);
    *(s + max) = '\0';
    fprintf(stderr, "%s\t%d\t%s\n", hash_db->hdb_name, i, s); 
    /* null-terminated */
  }
  return i;
}
int hash_put (HASH_DB *hash_db, const char *key,  int keylen) {
  unsigned long hval;
  HASH_ENTRY *he;
  const char *k;
  int id;

  k = md5hash(hash_db, key, &keylen);

  hval = hash (k, keylen, hash_db->hdb_size);
  he = &(hash_db->hdb_array[hval]);
  if (he->id != 0) {		/* ok, there was something */
    HASH_ENTRY *hen = he;
    do {
      if (hen->keylen == keylen &&
	  memcmp (hen->key, k, keylen) == 0) {
	return hen->id;	/* return id */
      }
      hash_db->hdb_cols++;
      he = hen;			/* store old */
    } while ((hen = hen->next));
    /* not found, but in linked list */
    if ((he->next = ALLOC_ONE_HASH_E()) == NULL) {
      perror("calloc");
      return 0;
    };
    he = he->next;
  }
  id = he->id = hash_db->hdb_index++; /* increment */
  he->keylen = keylen;
  memcpy (he->key, k, keylen);
#ifndef HASH_NORESCALE
  if ((id / hash_db->hdb_thres100) * HASH_TRESH_SCALE > hash_db->hdb_size)
    rescale (hash_db);
#endif
  return id;
}

int hash_get_id (HASH_DB *hash_db, const char *key,  int keylen) {
  unsigned long hval;
  HASH_ENTRY *he;
  const char *k;

  k = md5hash(hash_db, key, &keylen);

  hval = hash (k, keylen, hash_db->hdb_size);
  he = &(hash_db->hdb_array[hval]);

  if (he->id != 0) {		/* ok, there was something */
    HASH_ENTRY *hen = he;
    do {
      if (hen->keylen == keylen &&
	  memcmp (hen->key, k, keylen) == 0) {
	return hen->id;	/* return id */
      }
    } while ((hen = hen->next));
  }
  return 0;			/* not found */
}
void hash_stat(HASH_DB *hash_db) {
  int many = 0;
  int occupied = 0;
  int i;
  for (i = 0; i < hash_db->hdb_size; i++) {
    if (hash_db->hdb_array[i].id > 0) {
      occupied++;
      if (hash_db->hdb_array[i].next != NULL) 
	many++;
    }
  }
  fprintf (stderr,
	   "Hash statistics for %s: %d occupied, %d with more than one\n",
	   hash_db->hdb_name, occupied, many);
  fprintf (stderr,"Elements: %d\nCollisions: %d\n", hash_db->hdb_index - 1, 
	  hash_db->hdb_cols);
}
void hash_print (HASH_DB *hash_db) {
  unsigned long i;
  char buf[4 * MD5LEN + 1 + 8];	/* buffer for changes */
  for (i = 0; i < hash_db->hdb_size; i++) {
    HASH_ENTRY *he;
    for (he = &(hash_db->hdb_array[i]);
	 he != NULL && he->id != 0; he = he->next) {
      char *p;
      if (he->keylen == HASHKEYLEN) {
	int j;
	p = buf;
	for (j = 0; j < HASHKEYLEN; j++) {
	  if (isprint(he->key[j]) && he->key[j] != '\\')
	    *(p++) = he->key[j];
	  else {
	    sprintf (p, "\\x%08x", (he->key[j]) << 24);
	    p += 4;
	  }
	}
	*p = '\0';
	p = buf;
      } else {
	p = he->key;
      }
      fprintf (stderr, "%d\t%s\n", he->id, p);
    }
  }
}

HASH_DB *mkhash(char **strs, char *name, int hash_size, int thres100) {
  HASH_DB *hdb;
  int i;

  if ((hdb = hash_create(name, hash_size, thres100)) == NULL) {
    perror("mkhash (hash_create)");
    exit(3);
  }
  for (i = 0; strs[i]; i++) 
    hash_put_print(hdb, strs[i], strlen(strs[i]));
  return hdb;
}
